import json
import re
import os
from itertools import product

from .config import CONFUSABLE_MAPPING_PATH, NON_NORMAL_ASCII_CHARS
from .utils import is_ascii


# read confusable mappings from file, build 2-way map of the pairs
with open(os.path.join(os.path.dirname(__file__), CONFUSABLE_MAPPING_PATH), "r") as mappings:
    CONFUSABLE_MAP = json.loads(mappings.readline())


def is_confusable(str1, str2):
    while str1 and str2:
        length1, length2 = 0, 0
        for index in range(len(str1), 0, -1):
            if str1[:index] in confusable_characters(str2[0]):
                length1 = index
                break
        for index in range(len(str2), 0, -1):
            if str2[:index] in confusable_characters(str1[0]):
                length2 = index
                break

        if not length1 and not length2:
            return False
        elif not length2 or length1 >= length2:
            str1 = str1[length1:]
            str2 = str2[1:]
        else:
            str1 = str1[1:]
            str2 = str2[length2:]
    return str1 == str2

def confusable_characters(char):
    return CONFUSABLE_MAP.get(char)

def confusable_regex(string, include_character_padding=False):
    space_regex = "[\*|_|~|`|-|\.]*" if include_character_padding else ''
    regex = "\\b" + space_regex
    for char in string:
        regex += "[" + "|".join(CONFUSABLE_MAP[char]) + "|" + char + "]" + space_regex
    regex += "\\b"

    return regex

def normalize(string, prioritize_alpha=False):
    normal_forms = set([""])
    for char in string:
        normalized_chars = []
        confusable_chars = confusable_characters(char)
        if not is_ascii(char) or not char.isalpha():
            for confusable in confusable_chars:
                if prioritize_alpha:
                    if ((char.isalpha() and confusable.isalpha() and is_ascii(confusable)) or (not char.isalpha() and is_ascii(confusable))) and confusable not in NON_NORMAL_ASCII_CHARS:
                        normal = confusable
                        if len(confusable) > 1:
                            normal = normalize(confusable)[0]
                        normalized_chars.append(normal)
                else:
                    if is_ascii(confusable) and confusable not in NON_NORMAL_ASCII_CHARS:
                        normal = confusable
                        if len(confusable) > 1:
                            normal = normalize(confusable)[0]
                        normalized_chars.append(normal)
        else:
            normalized_chars = [char]

        if len(normalized_chars) == 0:
            normalized_chars = [char]
        normal_forms = set([x[0]+x[1].lower() for x in list(product(normal_forms, normalized_chars))])
    return sorted(list(normal_forms))



import json
from unicodedata import normalize
import string
import os
from config import CUSTOM_CONFUSABLE_PATH, CONFUSABLES_PATH, CONFUSABLE_MAPPING_PATH, MAX_SIMILARITY_DEPTH

def _asciify(char):
    return normalize('NFD',char).encode('ascii', 'ignore').decode('ascii')

def _get_accented_characters(char):
    return [u for u in (chr(i) for i in range(65536)) if u != char and _asciify(u) == char]

def _get_confusable_chars(character, unicode_confusable_map, depth):
    mapped_chars = unicode_confusable_map[character]

    group = set([character])
    if depth <= MAX_SIMILARITY_DEPTH:
        for mapped_char in mapped_chars:
            group.update(_get_confusable_chars(mapped_char, unicode_confusable_map, depth + 1))
    return group

def parse_new_mapping_file():
    unicode_confusable_map = {}

    with open(os.path.join(os.path.dirname(__file__), CONFUSABLES_PATH), "r") as unicode_mappings:
        with open(os.path.join(os.path.dirname(__file__), CUSTOM_CONFUSABLE_PATH), "r") as custom_mappings:
            mappings = unicode_mappings.readlines()
            mappings.extend(custom_mappings)

            for mapping_line in mappings:
                if not mapping_line.strip() or mapping_line[0] == '#' or mapping_line[1] == '#':
                    continue

                mapping = mapping_line.split(";")[:2]
                str1 = chr(int(mapping[0].strip(), 16))
                mapping[1] = mapping[1].strip().split(" ")
                mapping[1] = [chr(int(x, 16)) for x in mapping[1]]
                str2 = "".join(mapping[1])

                if unicode_confusable_map.get(str1):
                    unicode_confusable_map[str1].add(str2)
                else:
                    unicode_confusable_map[str1] = set([str2])

                if unicode_confusable_map.get(str2):
                    unicode_confusable_map[str2].add(str1)
                else:
                    unicode_confusable_map[str2] = set([str1])

                if len(str1) == 1:
                    case_change = str1.lower() if str1.isupper() else str1.upper()
                    if case_change != str1:
                        unicode_confusable_map[str1].add(case_change)
                        if unicode_confusable_map.get(case_change) is not None:
                            unicode_confusable_map[case_change].add(str1)
                        else:
                            unicode_confusable_map[case_change] = set([str1])

                if len(str2) == 1:
                    case_change = str2.lower() if str2.isupper() else str2.upper()
                    if case_change != str2:
                        unicode_confusable_map[str2].add(case_change)
                        if unicode_confusable_map.get(case_change) is not None:
                            unicode_confusable_map[case_change].add(str2)
                        else:
                            unicode_confusable_map[case_change] = set([str2])

    for char in string.ascii_lowercase:
        accented = _get_accented_characters(char)
        unicode_confusable_map[char].update(accented)
        for accent in accented:
            if unicode_confusable_map.get(accent):
                unicode_confusable_map[accent].add(char)
            else:
                unicode_confusable_map[accent] = set([char])

    for char in string.ascii_uppercase:
        accented = _get_accented_characters(char)
        unicode_confusable_map[char].update(accented)
        for accent in accented:
            if unicode_confusable_map.get(accent):
                unicode_confusable_map[accent].add(char)
            else:
                unicode_confusable_map[accent] = set([char])

    CONFUSABLE_MAP = {}
    characters_to_map = list(unicode_confusable_map.keys())
    for character in list(unicode_confusable_map.keys()):
        char_group = _get_confusable_chars(character, unicode_confusable_map, 0)

        CONFUSABLE_MAP[character] = list(char_group)

    mapping_file = open(os.path.join(os.path.dirname(__file__), CONFUSABLE_MAPPING_PATH), "w")
    mapping_file.write(json.dumps(CONFUSABLE_MAP))
    mapping_file.close()

parse_new_mapping_file()